
from pyArango.document import Document
from cag.framework.annotator.element.orchestrator import PipeOrchestrator


class EmpathPipeOrchestrator(PipeOrchestrator):

    def create_node(self, empath_cat) -> Document:
        data = {"category": empath_cat}
        return self.upsert_node(self.node_name, data, alt_key="category")

    def create_edge(self, from_:Document, to_:Document,
                    count: int,
                    ratio: float,
                    token_position_lst,
                    token_lst) -> Document:
        return self.upsert_edge(self.edge_name,
                                from_,
                                to_,
                                edge_attrs={
                                        "count": count,
                                        "ratio": ratio,
                                        "token_position_lst":token_position_lst, # array of tuples [(start, end), (start, end)]
                                        "token_lst": token_lst,  # array of tuples [(start, end), (start, end)]
                                    }
                                )

    col = "empath"
    col_position = "empath_position"
    col_words = "empath_words"
    def save_annotations(self, annotated_texts: "[]"):
        for doc, context in annotated_texts:
            text_key = context["_key"]

            for category, count in doc._.empath_count.items():
                empath_node: Document = self.create_node(category)
                text_node: Document = self.get_document(self.annotated_node, {"_key": text_key})
                _: Document = self.create_edge(text_node, empath_node,
                                                      doc._.empath_count[category],
                                                      doc._.empath_ratio[category],
                                                      doc._.empath_positions[category],
                                                      doc._.empath_words[category])












import hashlib
import importlib
import json
import pkgutil
import urllib
from re import sub


def get_hash_from_str(str: str) -> str:
    """generate a sha256-16 byte int from your string to use as a key

    :param str: the string to hash
    :type str: str
    :return: the hasehed string
    :rtype: str
    """
    hashing_func = hashlib.sha256
    def str2int(s): return int(hashing_func(s.encode()).hexdigest(), 16)
    return str2int(str)


def encode_name(name: str) -> str:
    """Encodes any string in a readable key (length may be of an issue)

    :param name: the string to encode
    :type name: str
    :return: the encoded name
    :rtype: str
    """
    enc_n = urllib.parse.quote_plus(name)
    enc_n = enc_n.replace("~", "%7E")
    return enc_n.replace("%", ")@(")


def to_dictionary(obj: object) -> dict:
    """Deep-clone an object by jsonifying it

    :param obj: the object to deep-clone
    :type obj: object
    :return: the nested dicts
    :rtype: dict
    """
    return json.loads(json.dumps(obj, default=lambda o: vars(o)))


def camel_case(s) -> str:
    # source: https://www.w3resource.com/python-exercises/string/python-data-type-string-exercise-96.php
    s = sub(r"(_|-)+", " ", s).title().replace(" ", "")
    return ''.join([s[0].lower(), s[1:]])


def filter_dic(obj, fields=None) -> dict:
    dict_ = to_dictionary(obj)
    dict_ = {k: v for k, v in dict_.items() if (
        (fields is None or k in fields) and v is not None)}
    return dict_


def camel_nest_dict(dict_) -> dict:
    res = {}
    for k, v in dict_.items():
        # Check if value is of dict type
        if isinstance(v, dict):
            res[camel_case(k)] = camel_nest_dict(v)
        else:
            # If value is not dict type then yield the value
            if v is not None:
                res[camel_case(k)] = v
    return res

def load_module(module_name:str):
    return importlib.import_module(module_name)

def get_cls_from_path(path: str):
    module_name, class_name = path.rsplit(".", 1)

    cls = getattr(importlib.import_module(module_name), class_name)
    return cls, class_name


def load_sub_packages(package):
    prefix = package.__name__ + "."
    #print(f"prefix {prefix}")
    for importer, modname, ispkg in pkgutil.iter_modules(package.__path__, prefix):
        #print(f"Found submodule {modname} (is a package: {ispkg})" )
        module = load_module(modname)
        #print (f"Imported {module}")


from datetime import datetime

from pyArango.theExceptions import DocumentNotFoundError, SimpleQueryError

from cag.utils.config import Config, configuration

from cag.graph_elements.base_graph import *
from pyArango.collection import Document, Collection
import re
from typing import Any, Optional

from cag import logger



class Component(object):
    """The class from witch all more specialized components must derive."""

    """
    Structure of the subgraph produced by this component, to be specified in subclass.
    Edge definitions are dictrionaries of the following form:
    {
        'relation': REL_NAME, # Must match a class inheriting from Edge
        'from_collections': COL_NAME, # Must match a class inheriting from Collection
        'to_collections': COL_NAME # Must match a class inheriting from Collection
    }
    """
    _edge_definitions = []
    _base_edge_definitions = []
    _description = "No description"
    _name = "Component"

    def __init__(self, conf: Config = None):
        edges = self._base_edge_definitions+self._edge_definitions

        if conf is None:
            conf = configuration(use_global_conf=True)
        self.conf = conf
        self.database = conf.db
        self.graph_name = conf.graph
        if self.database.hasGraph(self.graph_name):
            self.graph : BaseGraph= self.database.graphs[self.graph_name]
            
        else:

            edge_def_arr = []
            for ed in edges:
                for col in [ed['relation']] + ed['from_collections'] + ed['to_collections']:
                    if not self.database.hasCollection(col):
                        self.database.createCollection(col)

                edge_def_arr.append(EdgeDefinition(ed['relation'],
                                                   fromCollections=ed['from_collections'],
                                                   toCollections=ed['to_collections']))
            if len(edge_def_arr) == 0:
                raise CreationError("You have to define an edge for your graph in the your graph creator")
            _ = type(self.graph_name, (BaseGraph,), {'_edgeDefinitions': edge_def_arr})
            self.graph: BaseGraph = self.database.createGraph(
                self.graph_name)

        self.graph.__class__ = BaseGraph
        self.arango_db = conf.arango_db

        # Setup graph structure
        for ed in edges:
            self.graph.update_graph_structure(ed['relation'],
                                              ed['from_collections'],
                                              ed['to_collections'],
                                              create_collections=True
                                              )

    def get_document(self, collectionName: str, data: "dict[str, Any]", alt_key: "str | []" = None) -> "Optional[Document]":
        """Gets the node if it exists.
        In case alt_key is provided, it queries the node based on the alt_key keys and values in the data dict.
        In case alt_key is not provided and _key is part of the data dict, it fetches the document based on it.
        Otherwise it returns None

        :param collectionName: On which collection to search on
        :type collectionName: str
        :param data: the dict where the keys should be present
        :type data: dict[str, Any]
        :param alt_key: an alternate key to search on, defaults to None
        :type alt_key: str|list[str], optional
        :return: the found document, if not exists: None
        :rtype: Document, optional
        """
        '''

        '''
        coll: Collection = self.database[collectionName]
        node = None
        try:
            if type(alt_key) == str:
                alt_key = [alt_key]

            if alt_key is None and '_key' in data.keys() and data['_key'] in coll:
                node: Document = coll.fetchDocument(data['_key'])
            elif alt_key is not None and all(x in data.keys() for x in alt_key):
                coll.ensureHashIndex(alt_key, unique=True)

                query = {k: v for k, v in data.items() if k in alt_key}

                resp = coll.fetchByExample(
                    query,
                    batchSize=1)
                if len(resp) > 0:
                    node: Document = resp[0]
            else:
                logger.debug("node does not exist - make sure you provide _key as part of data dict "
                             "or alt_key as a lst and part of the data dict")
        except (DocumentNotFoundError, SimpleQueryError) as e:
            logger.debug("Document was not found for data {} and node {} - message: {}".format(
                collectionName, str(data), e.message))
        except Exception as unknown_e:
            logger.error("An unknown error was thrown for data {} and node {} - message: {}".format(
                collectionName, str(data), str(unknown_e)))
        return node

    def upsert_node(self, collectionName: str, data: "dict[str, Any]", alt_key: "str | []" = None) -> Document:
        """Upsert an item in a collection based on a _key or any other property

        :param collectionName: the collection to work on
        :type collectionName: str
        :param data: a dictionary with your data
        :type data: dict[str, Any]
        :param alt_key: on which key the upsert should look for existing data (if there are multiple, 
        the first match will return and it combines all key into a fetch-by-example-query), defaults to None
        :type alt_key: str | [], optional
        :return: the upserted Document
        :rtype: Document
        """
        coll: Collection = self.database[collectionName]

        if 'timestamp' not in data.keys() or data['timestamp'] is None:
            data['timestamp'] = datetime.now().isoformat()

        node = None
        try:
            node: Document = self.get_document(collectionName, data, alt_key)
            if node is None:
                node = self.graph.createVertex(collectionName, data)
            else:
                logger.debug("updating existing node")
                for key, d in data.items():
                    node[key] = d
                node.save()
                node = coll[node._key]

        except Exception as e:
            logger.error(
                "An unknown error was thrown for data {} and node {} - message: {}".format(collectionName, str(data),
                                                                                             str(e)))
        return node

    def get_edge_attributes(self, relationName: str, from_doc: Document, to_doc: Document, edge_attrs={}, add_id=""):
        """Gets a edge

        :param relationName: which edge collection to create this edge on
        :type relationName: str
        :param from_doc: the document where this relation starts
        :type from_doc: Document
        :param to_doc: the document where this relation leads to
        :type to_doc: Document
        :param edge_attrs: attributes to add to edge (provenance), defaults to {}
        :type edge_attrs: dict, optional
        :param add_id: data to append the id (if the from-to relation may not be unique), defaults to ""
        :type add_id: str, optional
        :return: the  edge document if it exists
        :rtype: 'Dict[str|Any]'
        """
        from_key = re.sub("/", "-", from_doc._id)
        to_key = re.sub("/", "-", to_doc._id)
        add_id = re.sub("/", "-", add_id)
        add_id = f'-{add_id}' if len(add_id) > 0 else ''
        edge_key = f'{from_key}-{to_key}{add_id}'
        self.database[relationName].validatePrivate("_from", from_doc._id)
        self.database[relationName].validatePrivate("_to", to_doc._id)

        edge_dic = {'_key': edge_key, '_from': from_doc._id,
                    '_to': to_doc._id, **edge_attrs}
        return edge_dic

    def upsert_edge(self, relationName: str, from_doc: Document, to_doc: Document, edge_attrs={}, add_id=""):
        """Upsert a edge (will generate a synthetic key from the provided documents, can optionally add something to these keys and edges)

        :param relationName: which edge collection to create this edge on
        :type relationName: str
        :param from_doc: the document where this relation starts
        :type from_doc: Document
        :param to_doc: the document where this relation leads to
        :type to_doc: Document
        :param edge_attrs: attributes to add to edge (provenance), defaults to {}
        :type edge_attrs: dict, optional
        :param add_id: data to append the id (if the from-to relation may not be unique), defaults to ""
        :type add_id: str, optional
        :return: the upserted edge document
        :rtype: Document
        """

        data = self.get_edge_attributes(
            relationName, from_doc, to_doc, edge_attrs, add_id)

        coll: Collection = self.database[relationName]

        if 'timestamp' not in data.keys() or data['timestamp'] is None:
            data['timestamp'] = datetime.now().isoformat()

        edge = None
        try:
            edge: Document = self.get_document(relationName, data)
            if edge is None:
                edge = self.graph.createEdge(relationName, from_doc._id, to_doc._id, data)
            else:
                logger.debug("updating existing node")
                for key, d in data.items():
                    edge[key] = d
                edge.save()
                edge = coll[edge._key]

        except Exception as e:
            logger.error(
                "An unknown error was thrown for data {} and node {} - message: {}".format(relationName, str(data),
                                                                                             str(e)))


        return edge





from abc import ABC, abstractmethod
from datetime import datetime
from typing import Any

from cag.framework.component import Component
from pyArango.collection import Document
from pyArango.query import AQLQuery
from cag.utils.config import Config


class GenericAnnotator(ABC, Component):
    def __init__(self, query: str, run=False, params={}, conf: Config = None, fetch_args: "dict[str,Any]" = {}, filter_annotatable=False,
                 annotator_fieldname="_annotator_params"):
        """the base class to extend your annotator from

        :param query: an arango query valid on your DB, returing the `annotator_fieldname` on the root-elements as a field from the docs to update
        :type query: str
        :param run: whether to run the analyzer on init, defaults to False
        :type run: bool, optional
        :param params: the parameters to set/check for on update, defaults to {}
        :type params: dict, optional
        :param conf: the config+connection to use, defaults to None
        :type conf: Config, optional
        :param fetch_args: the AQLQuery parameters (batch_size, etc.), defaults to {}
        :type fetch_args: dict[str,Any], optional
        :param filter_annotatable: Whether to filter your annotations, defaults to False
        :type filter_annotatable: bool, optional
        :param annotator_fieldname: the field to set your params on and filter the data, defaults to "_annotator_params"
        :type annotator_fieldname: str, optional
        """
        super().__init__(conf)
        self.annotator_fieldname = annotator_fieldname
        self.query = query
        self.now = datetime.now()
        self.params = params
        self.filter_annotatable = filter_annotatable
        if fetch_args is not None:
            self.fetch_args = fetch_args
        else:
            self.fetch_args = {}
        data = None
        if self.query is not None:
            data = self.fetch_annotator_data()
        if run:
            self.update_graph(self.now, data)

    def fetch_annotator_data(self):
        if self.filter_annotatable:
            query_modified = f"""
            LET annotator_data=({self.query})
            FOR dp IN annotator_data
                FILTER dp.{self.annotator_fieldname} == NULL or dp.{self.annotator_fieldname}!=@params
                RETURN dp
            """
            return self.database.AQLQuery(query_modified, bindVars={'params': self.params}, **self.fetch_args)
        else:
            return self.database.AQLQuery(self.query, **self.fetch_args)

    def complete_annotation(self, doc: Document):
        return self.upsert_node(doc.collection.name, doc.getStore())

    def upsert_node(self, collectionName: str, data: "dict[str, Any]", alt_key: "str | []" = None) -> Document:
        data[self.annotator_fieldname] = self.params
        return super().upsert_node(collectionName, data, alt_key)

    @abstractmethod
    def update_graph(self, timestamp, data: AQLQuery):
        g = self.graph

    def query_count(self)->int:
        """Get the total count of the query provided

        :return: the count of documents that *would* be returned
        :rtype: int
        """        """"""
        if self.query is None:
            return -1
        count_query = self.database.AQLQuery(f"""
LET query=({self.query})
RETURN COUNT(query)
        """, rawResults=True)
        return count_query[0]

    def load_page(self, offset: int = 0, limit: int = 500) -> AQLQuery:
        """Load a single page from the provided query

        :param offset: the offset to keep, defaults to 0
        :type offset: int, optional
        :param limit: the limit of the page, defaults to 500
        :type limit: int, optional
        :return: the executed query with the page (batchSize=limit, rawResults=True)
        :rtype: AQLQuery
        """""""""
        if self.query is None:
            return -1
        page_query = self.database.AQLQuery(f"""
LET query=({self.query})
FOR d IN query
    LIMIT @offset,@limit
    RETURN d
        """, bindVars={'offset': offset, 'limit': limit}, rawResults=True, batch_size=limit)
        return page_query



from ..test_nodes import *
from cag.utils.config import configuration
from ..text_db_fixture import *

class TestGCUpserts:
    def test_upsert(self):
        config = configuration(graph='SampleGraph', database=whoami())
        config_2 = configuration(graph='SampleGraph', database=whoami())
        gc = AnyGraphCreator("", config)
        gc.upsert_node(AnyGraphCreator._ANY_DATASET_NODE_NAME, {
            '_key': 'test_1',
            'name': 'before_upsert'
        })
        vert = config.arango_db[AnyGraphCreator._ANY_DATASET_NODE_NAME]['test_1']
        assert vert is not None
        assert vert['name'] == 'before_upsert'

        gc.upsert_node(AnyGraphCreator._ANY_DATASET_NODE_NAME, {
            '_key': 'test_1',
            'name': 'after_upsert'
        })
        # other connection, should still be updated
        vert = config_2.arango_db[AnyGraphCreator._ANY_DATASET_NODE_NAME]['test_1']
        assert vert is not None
        assert vert['name'] == 'after_upsert'
        


        # only update the correct doc
        gc.upsert_node(AnyGraphCreator._ANY_DATASET_NODE_NAME, {
            '_key': 'test_1',
            'name': 'before_insert'
        })
        vert = config.arango_db[AnyGraphCreator._ANY_DATASET_NODE_NAME]['test_1']
        assert vert is not None
        assert vert['name'] == 'before_insert'

        gc.upsert_node(AnyGraphCreator._ANY_DATASET_NODE_NAME, {
            '_key': 'test_2',
            'name': 'after_insert'
        })
        # other connection, should still be updated
        vert = config_2.arango_db[AnyGraphCreator._ANY_DATASET_NODE_NAME]['test_1']
        assert vert is not None
        assert vert['name'] == 'before_insert'
        vert = config_2.arango_db[AnyGraphCreator._ANY_DATASET_NODE_NAME]['test_2']
        assert vert is not None
        assert vert['name'] == 'after_insert'


    def test_upsert_alt_key(self):
        config = configuration(graph='SampleGraph', database=whoami())
        config_2 = configuration(graph='SampleGraph', database=whoami())
        gc = AnyGraphCreator("", config)
        gc.upsert_node(AnyGraphCreator._ANY_DATASET_NODE_NAME, {
            '_key': 'test_1',
            'key_alt': 'test_alt',
            'name': 'before_upsert'
        })
        vert = config.arango_db[AnyGraphCreator._ANY_DATASET_NODE_NAME]['test_1']
        assert vert is not None
        assert vert['name'] == 'before_upsert'
        gc.upsert_node(AnyGraphCreator._ANY_DATASET_NODE_NAME, {
            'key_alt': 'test_alt',
            'name': 'after_upsert'
        }, alt_key='key_alt')
        # other connection, should still be updated (on the alt key)
        vert = config_2.arango_db[AnyGraphCreator._ANY_DATASET_NODE_NAME]['test_1']
        assert vert is not None
        assert vert['name'] == 'after_upsert'



import dataclasses
from os import getenv
from pyArango.connection import *
from arango import ArangoClient

from arango.http import DefaultHTTPClient


@dataclasses.dataclass
class Config:

    url: "str | None" = None
    user: "str | None" = None
    password: "str | None" = None
    database: "str | None" = None
    graph: "str | None" = None
    autoconnect: bool = True
    timeout: int = 120

    def __post_init__(self):
        if self.url is None:
            self.url = getenv("ARANGO_URL", "http://127.0.0.1:8529")
        if self.user is None:
            self.user = getenv("ARANGO_USER", "root")
        if self.password is None:
            self.password = getenv("ARANGO_PW", "")
        if self.database is None:
            self.database = getenv("ARANGO_DB",  "_system")
        if self.graph is None:
            self.graph = getenv("ARANGO_GRAPH", "GenericGraph")
        if self.autoconnect:
            self.__connect()

    def __connect(self):
        # https://stackoverflow.com/questions/71838934/arangodb-read-timed-out-read-timeout-60
        class ArangoHTTPClient(DefaultHTTPClient):
            REQUEST_TIMEOUT = self.timeout  # Set the timeout you want in seconds here

        self.db: Database = None
        self.__connection = Connection(self.url, self.user, self.password)
        self.arango_client = ArangoClient(
            self.url, http_client=ArangoHTTPClient())
        if self.__connection.hasDatabase(self.database):
            self.db = self.__connection[self.database]
        else:
            self.db: Database = self.__connection.createDatabase(
                self.database)

        self.arango_db = self.arango_client.db(name=self.database,
                                               username=self.user,
                                               password=self.password)


global_conf = None


def configuration(url: "str | None" = None,
                  user: "str | None" = None,
                  password: "str | None" = None,
                  database: "str | None" = None,
                  graph: "str | None" = None, connect=True, use_global_conf=False, timeout=120) -> Config:
    """Start a new conenction using the provided config

    :param url: the adress of the ArangoDB, defaults to "http://127.0.0.1:8529"
    :type url: _type_, optional
    :param user: username, defaults to "root"
    :type user: str | None, optional
    :param password: password, defaults to ""
    :type password: str | None, optional
    :param database: database, will be created if it does not exist, defaults to "_system"
    :type database: str | None, optional
    :param graph: which graph to work on (must exist as class), defaults to "GenericGraph"
    :type graph: str | None, optional
    :param connect: whether to automatically connect, can be done later, defaults to True
    :type connect: bool, optional
    :param use_global_conf: if you want re-use one global config, defaults to False
    :type use_global_conf: bool, optional 
    :return: the connected config
    :rtype: Config
    """
    global global_conf
    if use_global_conf:
        if global_conf is not None:
            return global_conf
    conf = Config(url, user, password, database, graph,
                  autoconnect=connect, timeout=timeout)
    if use_global_conf:
        global_conf = conf
    return conf




#### HELPERS ####
from collections import defaultdict

import spacy
from spacy_arguing_lexicon import ArguingLexiconParser
from empath import Empath
from empath import helpers
from spacy.language import Language
from spacy.tokens import Token, Span, Doc


def set_extension_(ext, default = None):
    if not Doc.has_extension(ext):
        Doc.set_extension(ext, default=default)


### END OF Helpers ####



### Pipes ###

class EmpathExtended(Empath):
    def __init__(self):
        super().__init__()

    def analyze(self,doc,categories=None,tokenizer="default",normalize=False):
        if isinstance(doc,list):
            doc = "\n".join(doc)
        if tokenizer == "default":
            tokenizer = helpers.default_tokenizer
        elif tokenizer == "bigrams":
            tokenizer = helpers.bigram_tokenizer
        if not hasattr(tokenizer,"__call__"):
            raise Exception("invalid tokenizer")
        if not categories:
            categories = self.cats.keys()
        invcats = defaultdict(list)
        for k in categories:
           for t in self.cats[k]: invcats[t].append(k)
        count = {}
        ratio = {}
        tokens = 0.0
        for cat in categories: count[cat] = 0.0
        words_position = {}
        words = {}
        for tk in tokenizer(doc):
            tokens += 1.0
            for cat in invcats[tk]:
                count[cat]+=1.0
                if cat not in words_position.keys():
                    words_position[cat] = [(i, (i+len(tk))) for i in range(len(doc)) if doc.startswith(tk, i)]
                    words[cat] = []
                words[cat].append(tk)

        #if normalize:
        for cat in count.keys():
            if tokens == 0:
                return None
            else:
                ratio[cat] = count[cat] / tokens
        return count, ratio, words_position, words


@Language.component("empath_component")
def empath_component_function(doc:Doc):
    lexicon = EmpathExtended()
    empath_dic, ratios, positions, words = lexicon.analyze(doc.text)

    col_count = "empath_count"
    col_ratio = "empath_ratio"
    col_position = "empath_positions"
    col_words = "empath_words"

    set_extension_(col_count, default=None)
    set_extension_(col_ratio, default=None)
    set_extension_(col_position, default=None)
    set_extension_(col_words, default=None)

    empath_caregories = [k for k,v in empath_dic.items() if v > 0.0]
    doc._.set(col_count,    {k:v for k,v in empath_dic.items() if k in empath_caregories})
    doc._.set(col_ratio,    {k:v for k,v in ratios.items() if k in empath_caregories})
    doc._.set(col_position, {k:v for k,v in positions.items() if k in empath_caregories})
    doc._.set(col_words,    {k:v for k,v in words.items() if k in empath_caregories})


    return doc

@Language.component("mpqa_counter")
def mpqa_counter_function(doc):
    arguments = list(doc._.arguments.get_argument_spans_and_matches())

    total_arg_words = 0

    for arg in arguments:
        arg_span = arg[0]
        label = arg_span.label_

        col = "{}_{}".format("mpqa", label)
        set_extension_(col, default = 0)

        doc._.set(col, doc._.get(col)+1)

        total_arg_words += arg_span.__len__()

    set_extension_("mpqa_token_ratio", default = 0.0)
    set_extension_("count_mpqa_args", default = 0)

    doc._.mpqa_token_ratio =round(float(total_arg_words) / float(doc.__len__()), 3)
    doc._.count_mpqa_args = len(arguments)

    return doc


@Language.factory("mpqa_parser")
def mpqa_parser(nlp, name):
    return ArguingLexiconParser(lang=nlp.lang)


# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'Corpus Analytics Graph'
copyright = '2022, DLR'
author = 'DLR'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    'sphinx.ext.autodoc']


# generate autosummary even if no references
autosummary_generate = True
autoclass_content="both"


templates_path = ['_templates']
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'alabaster'
html_static_path = ['_static']



import spacy

from spacy.language import Language
from spacy.tokens import Token, Span, Doc

def count_tokens_w_attr( doc, attr):
    return {doc.vocab[k].text: v for k, v in doc.count_by(attr).items()}

def get_content_pos( pos_tuple=('verb', 'noun', 'adjective', 'adverb')):
    pos_labels = []
    for label in spacy.load("en_core_web_sm").get_pipe("tagger").labels:
        if spacy.explain(label).lower().startswith(pos_tuple):
            pos_labels.append(label)
    return pos_labels

@Language.component("descriptor")
def descriptor_function(doc):
    CONTENT_POS = get_content_pos()
    if not Doc.has_extension("num_token"):
        Doc.set_extension("num_token", default=0)
    if not Doc.has_extension("num_content_tokens"):
        Doc.set_extension("num_content_tokens", default = 0)

    doc._.num_token = doc.__len__()
    doc._.num_content_tokens = len([token for token in doc if token.tag_ in CONTENT_POS])

    pos_count_dict = count_tokens_w_attr(doc, spacy.attrs["POS"])

    for pos_tag, count in pos_count_dict.items():
        pos_tag_name = "{}_{}".format("pos_num_", pos_tag)
        if not Doc.has_extension(pos_tag_name):
            Doc.set_extension(pos_tag_name, default=0)
        doc._.set(pos_tag_name, count)

    return doc


import logging
from pathlib import Path

__version__ = "0.25.0"

logs_folder_name = 'cag_logs'
# User home dir detection. Linux and Windows compatible
home_directory = Path.home()
logs_folder = home_directory.joinpath(logs_folder_name)
if not logs_folder.exists():
    logs_folder.mkdir()

# Create a custom logger
global logger
logger = logging.getLogger(__name__)

# Create handlers
console_handler = logging.StreamHandler()
# console_handler.setLevel(logging.INFO)
console_format = logging.Formatter('%(name)-12s %(levelname)-8s %(message)s')
console_handler.setFormatter(console_format)
logger.addHandler(console_handler)

f_handler = logging.FileHandler(f"{logs_folder}/logs.txt")
# f_handler.setLevel(logging.INFO)
file_format = logging.Formatter('%(asctime)s %(name)-12s %(levelname)-8s %(message)s',
                                datefmt='%d.%m.%Y %H:%M')
f_handler.setFormatter(file_format)
logger.addHandler(f_handler)

logger.setLevel(logging.INFO)



from pyArango.collection import Edges, Field

"""GENREAL EDGE DEFINITION WITH TIMESTAMP"""
class GenericEdge(Edges):
    _fields = {
        'timestamp': Field()
    }

"""GENERAL GRAPH EDGES"""
class BelongsTo(GenericEdge):
    _fields = GenericEdge._fields

class RefersTo(GenericEdge):
    _fields = GenericEdge._fields

class IsIdenticalTo(GenericEdge):
    _fields = GenericEdge._fields

class HasAuthor(GenericEdge):
    _fields = GenericEdge._fields






import cag.graph_elements.nodes as parent_nodes
from cag.graph_elements.relations import GenericEdge
from pyArango.collection import  Field

"""OOS NODES RELEVANT FOR ANNOTATIONS"""


## NAMED ENTITY ANNOTATOR

class NamedEntityNode(parent_nodes.GenericNode):
    """ A class to define a Named Entity node in arangodb - This is used by pyarango to create the Collection """
    _fields = {
        "type": parent_nodes.Field(),
        "name": parent_nodes.Field(),
    }

    def __init__(self, database, jsonData):
        super().__init__(database, jsonData)
        self.ensureFulltextIndex(["name"], name="fti_annotator_ner")
        self.ensurePersistentIndex(["name", "type"], unique=True)


class EmpathNode(parent_nodes.GenericNode):
    """ A class to define a Named Entity node in arangodb - This is used by pyarango to create the Collection """
    _fields = {
        "category": parent_nodes.Field(),
    }

    def __init__(self, database, jsonData):
        super().__init__(database, jsonData)
        self.ensureFulltextIndex(["category"], name="fti_annotator_empath")
        self.ensurePersistentIndex(["category"], unique=True)

class HasAnnotation(GenericEdge):
    """ A class to define an annotation edge in arangodb - This is used by pyarango to create the Collection """

    _fields = {
        "count": Field(),
        "ratio": Field(),
        "token_position_lst": Field(), # array of tuples [(start, end), (start, end)]
        "token_lst": Field(),  # array of tuples [(start, end), (start, end)]
    }

    def __init__(self, database, jsonData):
        super().__init__(database, jsonData)
        self.ensurePersistentIndex(["_from", "_to"], unique=True)#, deduplicate=True)



from typing import ClassVar, List

from pyArango.document import Document
import pyArango
from cag.utils.config import Config
from cag.framework.annotator.element.orchestrator import PipeOrchestrator


class NamedEntityPipeOrchestrator(PipeOrchestrator):

    def create_node(self, ner_txt, ner_type) -> pyArango.document.Document:
        data = {"name": ner_txt, "type": ner_type}
        return self.upsert_node(self.node_name, data, alt_key=["name", "type"])

    def create_edge(self, _from: Document, _to: Document, entity) -> Document:
        position = (entity.start_char, entity.end_char)
        edge_dict: dict = self.get_edge_attributes(self.edge_name, _from, _to)
        edge = self.get_document(self.edge_name, edge_dict)

        lst_positions = [position]
        count: int = 1
        if edge is not None:
            if position not in edge.token_position_lst:
                if edge.token_position_lst is not None:
                    lst_positions.extend(edge.token_position_lst)
                if edge.count is not None:
                    count = count + edge.count

            else:
                return edge
        return self.upsert_edge(self.edge_name,
                                _from,
                                _to,
                                edge_attrs={"count": count,
                                            "token_position_lst": lst_positions
                                            }
                             )

    def save_annotations(self, annotated_texts:"[]"):
        for doc, context in annotated_texts:
            text_key = context["_key"]
            for ent in doc.ents:
                ner_txt = ent.text
                ner_type = ent.label_
                ner_node:Document = self.create_node(ner_txt, ner_type)
                text_node:Document = self.get_document(self.annotated_node, {"_key": text_key})
                _ :Document = self.create_edge(text_node, ner_node, ent)











import pytest

from cag.utils.config import configuration


@pytest.fixture(scope="session", autouse=True)
def database_cleanup():
    conn = configuration(database='_system')
    yield conn
    dbs = conn.arango_db.databases()
    for db in dbs:
        if 'test' in db:
            print("Deleting db", db)
            conn.arango_db.delete_database(db)



from arango import ViewGetError

from dataclasses import dataclass, field
from typing import List, ClassVar
from arango.database import StandardDatabase
from cag.utils import utils
from cag.view_wrapper.link import Link


@dataclass()
class ViewProperties():
    '''
    Contains the default properties for arango views
    '''
    cleanupIntervalStep: int =0
    ## Default arango db values
    cleanup_interval_step: int  = 0
    commit_interval_msec: int  = 1000
    consolidation_interval_msec: int  = 0
    consolidation_policy:dict  = field(default_factory = lambda:{
        "type": "tier",
        "segments_min": 1,
        "segments_max": 10,
        "segments_bytes_max": 5368709120,
        "segments_bytes_floor": 2097152,
        "min_score": 0
    })
    primary_sort_compression: str = "lz4"
    writebuffer_idle : int = 64
    writebuffer_active: int  = 0
    writebuffer_max_size: int  = 33554432



@dataclass
class View():
    name:str
    view_type:str = "arangosearch"
    properties: ViewProperties = field(default_factory = lambda:ViewProperties())
    links: List[Link] = field(default_factory = lambda:[])
    primary_sort : List = field(default_factory = lambda:[])
    stored_values: List = field(default_factory = lambda:[])

    _MANDATORY_FIELDS:ClassVar = ["name", "view_type"]

    def add_primary_sort(self, field_name:str, asc:bool):
        self.primary_sort.append({"field": field_name, "asc": asc})

    def add_stored_value(self, fields, compression ="lz4"):
        for field_name in fields:
            self.stored_values.append({"fields": [field_name], "compression": compression})

    def add_link(self, link: Link):
        self.links.append(link)

    def get_links_dict(self):
        dict_ = {}
        for l in self.links:
            dict_.update(l.summary())
        return dict_

    def get_properties(self) -> dict:
        mandatory_fields = View._MANDATORY_FIELDS
        result = utils.camel_nest_dict(utils.filter_dic(self))
        result["links"] = self.get_links_dict()
        for mandatory in mandatory_fields:
            del result[utils.camel_case(mandatory)]
        return result

    def summary(self) -> dict:
        result = utils.camel_nest_dict(utils.filter_dic(self))
        result["links"] = self.get_links_dict()
        return result

    def create(self, database: StandardDatabase):
        result = database.create_view(self.name,
                                      self.view_type,
                                      self.get_properties())
        return result

    def create_or_update(self, database: StandardDatabase):
        try:
            database.view(self.name)
            result = database.update_view(self.name,
                                 self.get_properties())
        except ViewGetError:
            result = database.create_view(self.name,
                                          self.view_type,
                                          self.get_properties())
        return result



from abc import ABC, abstractmethod

from cag import logger

from cag.utils.config import Config

from cag.framework.component import Component
from datetime import datetime


class GraphCreatorBase(ABC, Component):

    #### Constant NODE and Edge NAMES ####
    _TEXT_NODE_NAME = 'TextNode'
    _WEB_RESOURCE_NODE_NAME = 'WebResource'
    _AUTHOR_NODE_NAME = 'Author'
    _CORPUS_NODE_NAME = 'Corpus'
    _IMAGE_NODE_NAME = 'ImageNode'
    _DATA_NODE_NAME = 'DataNode'
    _KEY_TERM_NODE_NAME = 'KeyTerm'
    _ABSTRACT_NODE_NAME = 'Abstract'

    _BELONGS_TO_RELATION_NAME = 'BelongsTo'
    _REFERS_TO_RELATION_NAME = 'RefersTo'
    _HAS_AUTHOR_RELATION_NAME = 'HasAuthor'
    _PUBLISHED_AT_RELATION_NAME = 'PublishedAt'
    _EDGE_ABSTRACT_TEXT = 'AbstractText'
    _EDGE_TEXT_TERM = 'TextTerm'
    _EDGE_KEYTERM_RELATION = 'KeyTermRelation'

    _COMPONENT_NAME = 'Graph Creator'
    _base_edge_definitions = [
        {
            'relation': _EDGE_ABSTRACT_TEXT,
            'from_collections': [_ABSTRACT_NODE_NAME],
            'to_collections': [_TEXT_NODE_NAME]
        },
        {
            'relation': _EDGE_TEXT_TERM,
            'from_collections': [_TEXT_NODE_NAME],
            'to_collections': [_KEY_TERM_NODE_NAME]
        },
        {
            'relation': _EDGE_KEYTERM_RELATION,
            'from_collections': [_KEY_TERM_NODE_NAME],
            'to_collections': [_KEY_TERM_NODE_NAME]
        }
    ]

    def __init__(self, corpus_file_or_dir: str, conf: Config = None, initialize=False, load_generic_graph=True):
        """Creates a graph and provides some general helper methods

        :param corpus_file_or_dir: where your update_graph-method can find it's data 
        :type corpus_file_or_dir: str
        :param conf: config+connection to DB, defaults to None
        :type conf: Config, optional
        :param initialize: whether to initaialze the DB using init_graph(), defaults to False
        :type initialize: bool, optional
        :type load_generic_graph: bool, optional whether to load the predefined nodes and relations
        """
        if not load_generic_graph:
            self._base_edge_definitions = []
        super().__init__(conf)
        self.corpus_file_or_dir = corpus_file_or_dir
        self.now = datetime.now()
        if initialize:
            self.init_graph()

    @abstractmethod
    def init_graph(self):
        g = self.graph

    def update_graph(self, timestamp):
        g = self.graph

    ######### END OF HELPERS ###########################################################

    ######### Generic func to create vertices ##########################################

    def create_corpus_node(self, key, name, type, desc, created_on, timestamp=None):

        dict_ = {
            "_key": key,
            "name": name,
            "type": type,
            "description": desc,
            "created_on": created_on,
            "timestamp": timestamp
        }
        corpus = self.upsert_node(GraphCreatorBase._CORPUS_NODE_NAME, dict_)
        return corpus

    def create_text_node(self, text, timestamp=None):

        dict_ = {
            "text": text,
            "timestamp": timestamp
        }

        txt = self.upsert_node(
            GraphCreatorBase._TEXT_NODE_NAME, dict_, alt_key="text")
        return txt

    def create_image_node(self, url, timestamp=None):

        dict_ = {
            "url": url,
            "timestamp": timestamp
        }

        img = self.upsert_node(
            GraphCreatorBase._IMAGE_NODE_NAME,  dict_, alt_key="url")
        return img

    def create_author_node(self, author_name, timestamp=None):

        dict_ = {
            "name": author_name,
            "timestamp": timestamp
        }

        author = self.upsert_node(
            GraphCreatorBase._AUTHOR_NODE_NAME, dict_, alt_key="name")

        return author

    def create_web_resource_node(self, url, timestamp=None):

        dict_ = {
            "url": url,
            "timestamp": timestamp
        }

        web_resource = self.upsert_node(
            GraphCreatorBase._WEB_RESOURCE_NODE_NAME, dict_, alt_key="url")
        return web_resource
